// All the aliases accept by SyrusLang
const ALIASES = {
	// SyrusLang
	"define variable": "set variable",
	"set @tracking_resolution": "set tracking_resolution",
	"set $io.out": "set out",
	"enable counters": "start counters",
	"disable counters": "stop counters",
	// GPS
	"set gnss": "exec apx-gps",
	// Power Save
	"get power_save mode": "exec apx-psm mode",
	"set power_save mode": "exec apx-psm mode",
	"set power_save immediate": "exec apx-psm immediate",
	"get power_save": "exec apx-psm",
	// SMS
	"send sms": "exec apx-mdm send_sms",
	// Calls
	"start call": "exec apx-mdm start_call",
	"end call": "exec apx-mdm end_call",
	// Speak
	"speak lang=es": "exec apx-tts es-ES",
	"speak lang=en": "exec apx-tts en-US",
	"speak lang=fr": "exec apx-tts fr-FR",
	"speak es": "exec apx-tts es-ES",
	"speak en": "exec apx-tts en-US",
	"speak fr": "exec apx-tts fr-FR",
	"speak": "exec apx-tts en-US",
	// Bluetooth
	"bluetooth": "exec apx-bt",
	// apx-imu
	"get accelerometer all": "exec apx-imu config_list",
	"get accelerometer": "exec apx-imu",
	"set accelerometer": "exec apx-imu",
	// IButton
	"add ibutton": "exec apx-onewire ibutton add",
	"remove ibutton": "exec apx-onewire ibutton remove",
	"remove ibuttons": "exec apx-onewire ibutton remove_all",
	"get ibutton": "exec apx-onewire ibutton get",
	"get ibuttons": "exec apx-onewire ibutton get_all",
	// Temperature
	"add temperature_sensor": "exec apx-onewire temperature add",
	"remove temperature_sensor": "exec apx-onewire temperature remove",
	"remove temperature_sensors": "exec apx-onewire temperature remove_all",
	"get temperature_sensor": "exec apx-onewire temperature get",
	"get temperature_sensors": "exec apx-onewire temperature get_all",
	"get temperature": "exec apx-onewire temperature get_last",
	// Satcom
	"clear satcom buffer": "exec apx-serial modem clear_buffer",
	"set satcom buffer_size": "exec apx-serial modem buffer_size",
	"get satcom buffer_size": "exec apx-serial modem buffer_size",
	// Fatigue sensor
	"get fatigue_sensor state": "exec apx-serial fatigue_sensor state",
	"set fatigue_sensor sensitivity": "exec apx-serial fatigue_sensor sensitivity",
	"set fatigue_sensor speaker_volume": "exec apx-serial fatigue_sensor speaker_volume",
	"set fatigue_sensor minimum_speed": "exec apx-serial fatigue_sensor min_speed",
	"set fatigue_sensor speeding_alarm": "exec apx-serial fatigue_sensor speeding",
	"set fatigue_sensor buffer_size": "exec apx-serial fatigue_sensor max_photos",
	"set fatigue_sensor autoupload": "exec apx-serial fatigue_sensor auto_upload",
	"capture fatigue_sensor photo": "exec apx-serial fatigue_sensor capture_photo",
	"clear fatigue_sensor buffer": "exec apx-serial fatigue_sensor delete_photos",
	// MDT
	"set mdt": "exec apx-serial mdt settings",
	"send mdt message": "exec apx-serial mdt send",
	// "upload fatigue_sensor photos": "apx-serial fatigue_sensor upload",
	// Modem
	"set airplane_mode": "exec apx-mdm airplane_mode",
	"set sim_switch_interval": "exec apx-mdm sim_sw_interval",
	// Phonebook
	"add phone": "exec apx-mdm phone_book add",
	"remove phone": "exec apx-mdm phone_book remove",
	"set call_validation": "exec apx-mdm call_validation",
	"set sms_validation": "exec apx-mdm sms_validation",
};
// All the valids commands must be here so this can pre check if the command is no here there is a mistake on the config
const COMMANDS = [
	"exec",
	"define destination",
	"define tracking_resolution",
	"define signal",
	"define group",
	"define fieldset",
	"define event",
	"define action",
	"define geofence",
	"define counters",
	"define timer",
	"set value",
	"set namespace",
	"set destinations",
	"set variable",
	"set @signal_evaluation_cycle",
	"set tracking_resolution",
	"set counters",
	"set @ecu_params_response",
	"set out",
	"start counters",
	"stop counters",
	"delete",
	"disable", "stop",
	"enable", "start",
	"FLUSH ALL",
	"get",
	"send event",
	"trigger action",
	"fire signal",
];

// All the commands here must be added to the appends file because they modified permanently the config
const PERSISTENT_COMMANDS = [
	"define destination",
	"define tracking_resolution",
	"define signal",
	"define group",
	"define fieldset",
	"define event",
	"define action",
	"define geofence",
	"define counters",
	"set namespace",
	"set destinations",
	"set variable",
	"set @signal_evaluation_cycle",
	"set tracking_resolution",
	"set @ecu_params_response"
];

// list of arguments passed on the syruslang line that must be converted to units ej: min_duration=10min, speed_threshold=20kph
const TO_UNITS = [
	"min_duration",
	"odometer",
	"radius",
	"speed_threshold",
	"rpm_threshold",
	"begin_idle_time",
	"ignition_time",
	"idle_time",
	"over_speed",
	"over_rpm",
	"hard_brakes",
	"harsh_fwd_acceleration"
];

// help function to get degree on tracking_resolutions
function getDegree(line) {
	let regex = /(\d{1,5}(deg|rad))/gm;
	let resp = regex.exec(line);
	if (resp && resp[0]) return resp[0].trim();
	else return null;
}

// help function to get distance on tracking_resolutions
function getDistance(line) {
	let regex = /(\d{1,5}(mts|km|kms|mile))/gm;
	let resp = regex.exec(line);
	if (resp && resp[0]) return resp[0].trim();
	else return null;
}

// help function to get time on tracking_resolutions
function getTime(line) {
	let regex = /(\d{1,5}(hr|min|sec|s|h|m+$|m[^t]))/gm;
	let resp = regex.exec(line);
	if (resp && resp[0] && resp[0].endsWith("s")) console.warn(`${resp[0]} is now deprecated you should use 'sec' as unit`);
	if (resp && resp[0] && resp[0].endsWith("m")) console.warn(`${resp[0]} is now deprecated you should use 'min' as unit`);
	if (resp && resp[0] && resp[0].endsWith("h")) console.warn(`${resp[0]} is now deprecated you should use 'hr' as unit`);
	if (resp && resp[0]) return resp[0].trim();
	else return null;
}

// helper function to get csv pair like $ecu:$ecu,foo:bar as { "$ecu": "$ecu", "foo": "bar"}
// easy to handle for javascript, ej: fieldset use this config, also headers for mqtt and http
function getAsCSVPair(val) {
	var results = {};
	var pairs = val.split(",");
	pairs.forEach(txt => {
		results[txt.split(":")[0]] = txt.split(":")[1];
	});
	return results;
}

class SyrusLang {
	constructor() {}

	/**
	 * replace in syruslang Line aliases for equivalent
	 * @param {*} command syruslang line  config
	 */
	checkAliases(command) {
		for (const alias in ALIASES) {
			if (command.startsWith(alias)) return command.replace(alias, ALIASES[alias]);
		}
		return command;
	}

	/**
	 * function used to convert syruslang define geofence in geojson
	 * @param {*} latlngsString the latlngs part of the syruslang
	 * @param {*} radius the radius part of the syruslang
	 * @param {*} type the type of the geofence could be auto,Polygon or Circular
	 */
	toGeoJSON(latlngsString, radius, type = "auto") {
		let regexCircular = /^(\s?\-?\d{1,99}\.?\d{1,99}\,\-?\d{1,99}\.?\d{1,99}\s?){1,1}$/g;
		if (regexCircular.test(latlngsString)) {
			return {
				type: "Feature",
				geometry: {
					coordinates: latlngsString.split(",").map(l => Number(l)),
					radius: radius || 50,
					type: "Circle"
				},
				properties: {
					radius: radius
				}
			};
		}
		var regex = /^(\s?\-?\d{1,99}\.?\d{1,99}\,\-?\d{1,99}\.?\d{1,99}\,?){2,}(\,?\-?\d{1,99}\.?\d{1,99}\,\-?\d{1,99}\.?\d{1,99}){0,}\s?$/g;
		if (!regex.test(latlngsString)) {
			throw "invalid latlngs " + latlngsString;
		}
		var coords = latlngsString
			.trim()
			.split(" ")
			.map(latlng => {
				return latlng.split(",").map(l => Number(l));
			});
		var type = radius > 0 ? "LineString" : "Polygon";
		if (type == "Polygon") {
			var first = coords[0];
			var last = coords[coords.length - 1];
			if (first[0] !== last[0] || first[1] !== last[1]) {
				coords.push([first[0], first[1]]);
			}
		}
		return {
			type: "Feature",
			geometry: {
				coordinates: type == "LineString" ? coords : [[coords]],
				radius: radius,
				type: type
			},
			properties: {
				radius: radius
			}
		};
	}

	/**
	 * Entry point of this class where all the syruslang is received
	 * @param {*} text accepts Syruslang lines, could be a single command or the whole config
	 * @returns array of SyrusLang Json actions, ignores invalid commands
	 */
	process(text) {
		var lines = text.split(/\r?\n/);
		lines = lines.filter(c => {
			if (c.startsWith("#") || c == "") {
				return false;
			}
			return true;
		});
		var commands = [];
		lines.forEach(line => {
			var command = this._process(line.trim());
			if (command) {
				command.raw = line;
				commands.push(command);
			}
		});
		return commands.filter(v => !!v);
	}

	/**
	 * !Important Single line process of syruslang parser
	 * !All the important logic starts here
	 * @param {*} _line syrusLang line command
	 * @returns syruslang action object
	 */
	_process(_line) {
		let line = this.checkAliases(_line); // onvert alias to syruslang
		let command;
		// Checkign is a valid command
		for (const cmd of COMMANDS) {
			if (line.startsWith(cmd)) {
				command = cmd;
				break;
			}
		}
		if (!command) {
			console.warn("not supported: ", line);
			return;
		}
		// Getting all parameters from syruslang
		let args = this.getArguments(line);
		args.command = command;
		args.persistent = PERSISTENT_COMMANDS.includes(command); // if persisteng flag is set to append to the appends.syrus.conf

		// Specials
		// Has unix commands
		if (args.command == "exec") {
			return {
				command: "exec",
				argument: line.replace("exec", "")
			};
		}
		// Has syruslang command
		if (args.command == "define action") {
			let name = line.split(" ")[2];
			let signals = this.getArg(line, "trigger", false)
			if (signals) signals = signals.split(',')
			return {
				name,
				signals,
				command: "define action",
				persistent: true,
				rate: args.rate ? this.getRater(args.rate) : null,
				action: line
					.replace(`define action ${name}`, "")
					.replace(`trigger=${this.getArg(line, "trigger", false)}`, "")
					.replace(`rate=${this.getArg(line, "rate", false)}`, "")
					.trim()
			};
		}
		// design to delete the whole config
		if (args.command == "FLUSH ALL") {
			return {
				command: "FLUSH ALL"
			};
		}

		// Depending of the command the different argument must be processed
		if (args.command == "define destination") {
			let [, , name, protocol, url] = line
				.split(" ")
				.map(arg => arg.trim())
				.filter(m => m);
			args.name = name;
			args.protocolId = args.protocol;
			args.protocol = protocol;
			args.persistent = true;
			if (args.channelRx && !args.subscribe) {
				args.subscribe = args.channelRx;
				console.warn("channelRx is now deprecated you should use subscribe instead [mqtt destination]");
			}
			if (args.channelTx && !args.publish) {
				console.warn("channelTx is now deprecated you should use publish instead [mqtt destination]");
				args.publish = args.channelTx;
			}
			args.channel = url.split("://")[0];
			args.host = url.split("://")[1].split(":")[0];
			args.port = url.split("://")[1].split(":")[1] || 5000;
			if (args.headers) args.headers = getAsCSVPair(args.headers);
			if (args.params) args.params = getAsCSVPair(args.params);
		} else if (args.command == "define tracking_resolution") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			let heading = getDegree(line);
			let time = getTime(line);
			let distance = getDistance(line);
			args.name = name;
			args.persistent = true;
			args.time = !isNaN(parseInt(time)) ? this.toBasicUnit(time) : 0;
			args.heading = !isNaN(parseInt(heading)) ? this.toBasicUnit(heading) : 0;
			args.distance = !isNaN(parseInt(distance)) ? this.toBasicUnit(distance) : 0;
		} else if (args.command == "define signal") {
			let name = line.split(" ")[2];
			let [variable, operator, comparator] = line
				.replace(`define signal ${name}`, "")
				.replace(`min_duration=${this.getArg(line, "min_duration", false)}`, "")
				.trim()
				.split(" ")
				.map(arg => arg.trim());
			args.name = name;
			args.persistent = true;
			args.variable = variable;
			args.operator = operator;
			args.comparator = comparator;
		} else if (args.command == "define group") {
			args.persistent = true;
			args.name = line.split(" ")[2];
			args.meta = line.split(" ")[3];
		} else if (args.command == "define fieldset") {
			args.name = line.split(" ")[2];
			args.fields = this.fieldset(line);
			args.persistent = true;
		} else if (args.command == "define event") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.name = name;
			args.code = args.code || "01";
			args.label = args.label || "trckpnt";
			args.group = args.group || "default";
			args.fieldset = args.fieldset || "default";
			args.ack = args.ack || "imei";
			args.retain = `${args.retain}` == "true";
			args.destinations = this.getArg(line, "destinations", false)
			if (args.destinations) args.destinations = args.destinations.split(",")
			args.signals = this.getArg(line, "trigger", false);
			if (args.signals) args.signals = args.signals.split(",")
			args.rate = args.rate ? this.getRater(args.rate) : null;
			args.photo = this.getArg(line, "photo", false)
		} else if (args.command == "define geofence") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
			args.group = this.getArg(line, "group", false)
			args.group = args.group ? args.groups.split(',') : []
			args.lngLats = line
				.replace(name, "")
				.replace("define geofence", "")
				.replace(`group=${args.group.join(",")}`, "")
				.replace(`radius=${this.getArg(line, "radius", false)}`, "")
				.trim();
		} else if (args.command == "define counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
			args.forceSet = false;
		} else if (args.command == "define timer"){
			let [, , name] = line.split(" ", 3)
			args.name = name
			args.repeat = this.getArg(line, "repeat", false) || false
			args.enabled = this.getArg(line, "enabled", false) || false
			let duration = this.getArg(line, "duration", true)
			let delay = this.getArg(line, "delay", true)
			args.duration = !isNaN(parseInt(duration)) ? parseInt(duration) : 0
			args.delay = !isNaN(parseInt(delay)) ? parseInt(delay) : 0
			if (!args.duration){
				console.warn(`define timer ${name}: missing duration argument`)
			}
		} else if (args.command == "set namespace") {
			args.name = line.split(" ")[2];
		} else if (args.command == "set destinations") {
			args.destinations = line
				.split(" ")
				.reverse()[0]
				.split(",");
		} else if (args.command == "set variable") {
			let [, , name, value] = line.split(" ", 4);
			if (value.endsWith('"')) value = value.substring(2, value.length - 2);
			args.name = name;
			args.value = value;
		} else if (args.command == "set tracking_resolution") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			let heading = getDegree(line);
			let time = getTime(line);
			let distance = getDistance(line);
			args.name = line.split(" ")[2];
			args.time = !isNaN(parseInt(time)) ? this.toBasicUnit(time) : 0;
			args.heading = !isNaN(parseInt(heading)) ? this.toBasicUnit(heading) : 0;
			args.distance = !isNaN(parseInt(distance)) ? this.toBasicUnit(distance) : 0;
		} else if (args.command == "set counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
			args.forceSet = true;
		} else if (args.command == "set @ecu_params_response") {
			let [, key, val] = line.split(" ", 3);
			args.key = key;
			args.val = val;
		} else if (args.command == "set out") {
			let [, io, state] = line.split(" ");
			args.io = io.replace("$io.out", "").replace("out", "");
			args.state = state;
		} else if (args.command == "start counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
		} else if (args.command == "stop counters") {
			let [, , name] = line.split(" ", 3);
			args.name = name;
		} else if (args.command == "delete") {
			let [, type, name] = line.split(" ");
			args.name = name;
			args.type = type;
		} else if (args.command == "enable" || args.command == "start") {
			let [_, type, name] = line.split(" ");
			args.type = type;
			args.name = name;
			args.state = "enable";
		} else if (args.command == "disable" || args.command == "stop") {
			let [_, type, name] = line.split(" ");
			args.type = type;
			args.name = name;
			args.state = "disable";
		} else if (args.command == "get") {
			let [, type, name] = line.split(" ");
			args.type = type.replace("_value", "");
			args.variable_name = name;
			args.returnValue = type.endsWith("_value");
		} else if (args.command == "set value") {
			let [, , name, value] = line.split(" ");
			args.variable_name = name;
			args.value = value;
		} else if (args.command == "send event") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.name = name;
		} else if (args.command == "trigger action") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.name = name;
		} else if (args.command == "fire signal") {
			let [, , name] = line.split(" ").map(arg => arg.trim());
			args.signal_name = name;
		}
		return args;
	}

	/**
	 *  function to get single argument from a syruslang command fieldset="..."
	 * @param {*} line the line itself to extract the argument
	 * @param {*} variableName  the name of the argument
	 * @param {*} convert true if the argument must be convert to units like time,distance,rpm,speed, etc...
	 */
	getArg(line, variableName, convert = true) {
		let index = line.indexOf(`${variableName}=`);
		if (index == -1) return null
		let value = line.split(`${variableName}=`)[1]
		if (value.startsWith('"')) value = value.split('"')[1]
		else if (value.startsWith("'")) value = value.split("'")[1]
		else value = value.split(' ')[0]
		value = JSON.parse(`"${value}"`)
		if (convert) value = this.toBasicUnit(value);
		return value;
	}

	/**
	 * Spacial function to get the rater argument that it's composed in two parts: limit/time
	 * @param {*} val The values from the rater
	 */
	getRater(val) {
		let [count, time] = val.split("/");
		time = this.toBasicUnit(time);
		return {
			times: count,
			timeout: time
		};
	}

	/**
	 * Get all the arguments from a syruslang command
	 * @param {string} line the syruslang line command
	 * @returns An object with the values of the all arguments
	 */
	getArguments(line) {
		const regex = /\w{1,}=\S{1,}/gm;
		let m;
		let args = {};
		while ((m = regex.exec(line)) !== null) {
			// This is necessary to avoid infinite loops with zero-width matches
			if (m.index === regex.lastIndex) {
				regex.lastIndex++;
			}

			// The result can be accessed through the `m`-variable.
			m.forEach(match => {
				let [key] = match.split("=");
				let convert = TO_UNITS.includes(key);
				args[key] = this.getArg(line, key, convert);
			});
		}
		return args;
	}

	/**
	 * function to convert units in the units accepted by syrusjs example 1hr or 1min to ms that is expected for syrusjs
	 * @param {string} value to convert include the unit
	 */
	toBasicUnit(value) {
		if (!isNaN(parseInt(value))) {
			// Distance
			if (value.endsWith("km")) {
				value = parseInt(value) * 1000;
			} else if (value.endsWith("mts")) {
				value = parseInt(value) * 1;
			} else if (value.endsWith("mile")) {
				value = parseInt(value) * 1609;
			} else if (value.endsWith("ft")) {
				value = parseInt(parseInt(value) * 0.3048);
			}
			// Speed
			else if (value.endsWith("kph")) {
				value = parseInt(value) * 1;
			} else if (value.endsWith("mph")) {
				value = parseInt(value) * 1.60934;
			} else if (value.endsWith("m/s")) {
				value = parseInt(value) * 3.6;
			}
			// Time
			else if (value.endsWith("s") || value.endsWith("sec")) {
				value = parseInt(value) * 1000;
			} else if (value.endsWith("h") || value.endsWith("hr")) {
				value = parseInt(value) * 1000 * 3600;
			} else if (value.endsWith("m") || value.endsWith("min")) {
				value = parseInt(value) * 60 * 1000;
			}
			// Degree
			else if (value.endsWith("deg")) {
				value = parseInt(value);
			} else if (value.endsWith("rad")) {
				value = parseInt(parseInt(value) * (180 / Math.PI));
			}
			// Cycles
			else if (value.endsWith("hz")) {
				value = parseInt(value) * 60;
			} else if (value.endsWith("rpm")) {
				value = parseInt(value) * 1;
			}
			// default
			else {
				value = parseInt(value);
			}
		}
		return value;
	}

	/**
	 * helper function to process the fieldset argument
	 * @param {*} line the syruslang line command
	 * @returns the fields injavascript object
	 */
	fieldset(line) {
		var _fields = line.substring(line.indexOf(`fields=`) + 7).split(",");
		var fields = {};
		for (const field of _fields) {
			var [key, value] = field.split(":");
			fields[key] = value ? value : key;
		}
		return fields;
	}
}

module.exports = SyrusLang;
