// Protocols
const TCP = require("../channels/TCP");
const HTTP = require("../channels/HTTP");
const File = require("../channels/File");
const MQTT = require("../channels/MQTT");
const Satcom = require("../channels/Satcom");

// Channels
const TAIP = require("../protocols/TAIP");
const Json = require("../protocols/Json");
const { startCounters } = require("./counters");

module.exports = {
  /**
   * Creates a destination (which is basically a channel-protocol pair) with all the config passed in the syruslang
   *
   * @param {*} config all the parameters from syruslang to create the destination like protocol, port, host, name, etc
   */
  createDestination(config, callback) {
    var { name, protocol, channel, host, port, user, password, fieldset, headers, params } = config;
    var { ack, disabled } = config;
    var client, parser;

    if (!ack) ack = "seq"
    if (!"tcp,mqtt".includes(channel)) ack = "disabled"

    // function to procces syruslang commands from every destination (as TAIP receives  STX)
    // this is passed to every destination

    /** Channels Creation */
    if (channel == "tcp") {
      client = new TCP(config);
    } else if (channel == "file") {
      client = new File({
        host
      });
    } else if (channel == "http" || channel == "https") {
      client = new HTTP({
        name,
        headers,
        fieldset,
        user,
        password,
        host,
        port,
        params,
        protocol: channel,
        method: config.method,
        onCommand: callback
      });
    } else if (channel == "mqtt") {
      config.onCommand = callback;
      client = new MQTT(config);
    } else if (channel == "satcom"){
      config.onCommand = callback;
      client = new Satcom(config)
    }

    if (disabled) client.disable()

    /** Protocols Creation */
    if (protocol == "taip") {
      parser = new TAIP({
        client: client,
        buffer: [],
        onCommand: callback,
        ack: ack,
        fieldset: fieldset,
        name: name
      });
    } else if (protocol == "json") {
      parser = new Json({
        client: client,
        buffer: [],
        onCommand: callback,
        fieldset: fieldset,
        name: name,
        ack: ack
      });
    }

    parser.onCommand = callback;

    // The destination object allow easily verify is the destination exists and what event belong to
    return {
      client: client,
      parser: parser,
      protocol: protocol,
      channel: channel,
      host: host,
      port: port,
      user: user,
      password: password,
      // sendEvent is called for the main module everytime signals are up and the event must be fired
      sendEvent(data, event, fields, label, code) {
        this.parser.send(data, event, fields, label, code);
      },
      // callback called when the app is about to die, to make clean disconnection (as mqtt for example send the last will)
      disconnect() {
        client.disconnect();
      }
    };
  }
}
// TODO: delete Destination Point
